/*
    Session Sniffer v1.1
    written by Bartosz Bogacki <arkth@team.com.pl> '2001

    greetz: SigSegv, BugzPL, \"Olka Najlepsza\", ...
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#ifndef __USE_BSD
#define __USE_BSD
#endif
#include <sys/types.h>
#include <sys/socket.h>
#define __FAVOR_BSD
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <pwd.h>
#include <ctype.h>

#define PKBUFSIZE 4096

void
error (char *msg)
{
    perror (msg);
    exit (1);
}

void
usage ()
{
    printf ("Session Sniffer v1.1 ");
    printf ("written by Bartosz Bogacki <arkth@team.com.pl>\n\n");
    printf ("usage:\n");
    printf ("\tss [options][<arguments>]\n\n");
    printf ("options:\n");
    printf (" -h\t\t\tthis help screen :>\n");
    printf (" -s <hostname>\t\tshow payload of packets from/to this host\n");
    printf (" -p <port1,port2,...>\tshow payload of packets from/to this ports\n");
    printf (" -b <port1,port2,...>\tbypass this ports while sniffing\n");
    printf (" -e <expr1,expr2,...>\tmatch expressions to packets payload\n");
    printf ("\t\t\t[exprs are separated with \",\" (comma)]\n");
    printf ("\n");
    printf ("examples:\n");
    printf ("\t./ss\n");
    printf ("\t./ss -s shf.dhs.org -b 22,443,80\n");
    printf ("\t./ss -s dsi.pl -p 21,110\n");
    printf ("\t./ss -s rapid.black.pl -p 110 -e USER,PASS\n");
    printf ("\t./ss -e 'zabic prezydent,admin to dupa'\n");
    printf ("\n");
    exit (0);    
}

long
host2addr (char *host)
{
    long addr = 0;
    struct hostent *h = NULL;

    if ((addr = inet_addr (host)) < 0)
	if (h = gethostbyname (host)) 
	    memcpy ((char *)&addr, (char *)h->h_addr, 4);
	else error ("gethostbyname");
	
    return (addr);
}

int
main (int argc, char **argv)
{
    u_int16_t sport;
    u_int16_t dport;
    int sd;
    int i;
    int pksize;
    int opt;
    int expr_number = 0;
    u_int16_t bypass_number = 0;
    u_int16_t port_number = 0;
    int expr_ok;
    int bypass_ok;
    int port_ok;
    long host = 0;
    u_int16_t *bypass_array;
    u_int16_t *port_array;
    char *ptr;
    char pkbuf[PKBUFSIZE];
    char **expr_array;
    struct ip *ip;
    struct tcphdr *tcph;

/* parse options */
    while ((opt = getopt (argc, argv, "s:p:e:b:h")) != -1)
    {
	switch (opt)
	{
	    case 's': host = host2addr (optarg);
		      break;
	    case 'p': port_number = 1;
		      if ((optarg == NULL) ||
		           (bypass_number)) usage ();

		      for (i = 0; i < strlen (optarg) - 1; ++i)
		          if (optarg[i] == ',') ++port_number;
			
		      if ((port_array = 
		          (u_int16_t *)calloc (port_number + 1, sizeof (u_int16_t))) == NULL)
			      usage ();

		      for (i = 0; i < port_number; ++i)
		      {
		          port_array[i] = atoi (strsep (&optarg, ","));
		      }
		      break;		      
	    case 'e': expr_number = 1;
		      if (optarg == NULL) usage ();

		      for (i = 0; i < strlen (optarg) - 1; ++i)
		          if (optarg[i] == ',') ++expr_number;
			
		      if ((expr_array = 
		          (char **)calloc (expr_number + 1, sizeof (char *))) == NULL)
			      usage ();

		      for (i = 0; i < expr_number; ++i)
		      {
		          expr_array[i] = strdup (strsep (&optarg, ","));
			  if (!strcmp (expr_array[i], "")) usage ();
		      }
		      break;		      
	    case 'b': bypass_number = 1;
		      if ((optarg == NULL) ||
		          (port_number)) usage ();

		      for (i = 0; i < strlen (optarg) - 1; ++i)
		          if (optarg[i] == ',') ++bypass_number;
			
		      if ((bypass_array = 
		          (u_int16_t *)calloc (bypass_number + 1, sizeof (u_int16_t))) == NULL)
			      usage ();

		      for (i = 0; i < bypass_number; ++i)
		      {
		          bypass_array[i] = atoi (strsep (&optarg, ","));
		      }
		      break;		      
	    case 'h':
	    default: usage ();
	}
    }
    
    if ((sd = socket (PF_INET, SOCK_RAW, IPPROTO_TCP)) < 0) error ("socket");
    while (1)
    {
	memset (pkbuf, 0, sizeof (pkbuf));
	if ((pksize = read (sd, pkbuf, PKBUFSIZE)) < 0) continue;

	ip = (struct ip *)pkbuf;
	tcph = (struct tcphdr *)(pkbuf + (ip->ip_hl << 2));

/* check hosts */
	if ((host) && (host != *(long *)&ip->ip_src)) continue;
	sport = ntohs (tcph->th_sport);
	dport = ntohs (tcph->th_dport);

/* bypass lines without data... nah... */
	if ((tcph->th_flags != TH_PUSH) &&
	    (tcph->th_flags != (TH_PUSH|TH_ACK))) continue;

/* check ports */
	if (port_number)
	{
	    port_ok = 0;
	    for (i = 0; i < port_number; ++i)
	    {
		if ((port_array[i] == sport) ||
		    (port_array[i] == dport))
		    {
			port_ok = 1;
			break;
		    }
	    }
	    if (!port_ok) continue;
	}
	else if (bypass_number)
	{
	    bypass_ok = 0;
	    for (i = 0; i < bypass_number; ++i)
	    {
		if ((bypass_array[i] == sport) ||
		    (bypass_array[i] == dport))
		    {
			bypass_ok = 1;
			break;
		    }
	    }
	    if (bypass_ok) continue;
	}

/* set pointer to data */
	ptr = (char *)tcph + (tcph->th_off << 2);
	pksize -= ((tcph->th_off << 2) + (ip->ip_hl << 2));
/* check if expression match to data */
	expr_ok = 0;
	if (expr_number)
	{
	    for (i = 0; i < expr_number; ++i)
		if (memmem (ptr, pksize, expr_array[i], strlen (expr_array[i])))
		    {
			expr_ok = 1;
			break;
		    }
	    if (!expr_ok) continue;
	}

/* print header */
	printf ("\n[ %s:%d -> 0.0.0.0:%d ]\n",
	    inet_ntoa (*(struct in_addr *)&ip->ip_src),
	    sport, dport);

	for (i = 0; i < pksize; ++i) 
	{
	    if (isprint (ptr[i])) putc (ptr[i], stdout);
	    else if (ptr[i] == '\n') printf ("\n");
	    else if (ptr[i] == '\r') ;
	    else if (ptr[i] == '\b') printf ("^H");
	    else if (ptr[i] == '\t') printf ("\t");
	    else putc ('.', stdout);
	}
	fflush (NULL);
    }

/* never reached */
    exit (0);
}
